#include "pawn_lm3s3748_IO.h"

#define DIGITAL_IO_MAX 16

#include "inc/hw_memmap.h"
#include "inc/hw_types.h"
#include "driverlib/sysctl.h"
#include "driverlib/gpio.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"

static const unsigned long g_ulSYSCTL_PERIPH_GPIO[8] =
{
    SYSCTL_PERIPH_GPIOA, SYSCTL_PERIPH_GPIOB, SYSCTL_PERIPH_GPIOC, SYSCTL_PERIPH_GPIOD, 
    SYSCTL_PERIPH_GPIOE, SYSCTL_PERIPH_GPIOF, SYSCTL_PERIPH_GPIOG, SYSCTL_PERIPH_GPIOH
};

static const unsigned long g_ulGPIOBase[8] =
{
    GPIO_PORTA_BASE, GPIO_PORTB_BASE, GPIO_PORTC_BASE, GPIO_PORTD_BASE,
    GPIO_PORTE_BASE, GPIO_PORTF_BASE, GPIO_PORTG_BASE, GPIO_PORTH_BASE
};

unsigned char g_NumDigitalOut = 0;
unsigned char g_NumDigitalIn  = 0;

struct GPIO_PIN_INFO {
	unsigned long port;		/* |[g */
	unsigned char pin;		/* sԍ */
};

struct PIO_INFO {
	struct GPIO_PIN_INFO Pins[DIGITAL_IO_MAX];
};

struct PIO_INFO g_DO;
struct PIO_INFO g_DI;

int amx_putstr(const char* s);

static cell AMX_NATIVE_CALL n_PinTypeGPIOOutput(AMX *amx, const cell *params)
{
	unsigned long pf_code = ( unsigned long )params[1];
	unsigned char pin_code = ( unsigned char )( pf_code >> 24 );
	unsigned long ulPort;
	unsigned char ucPins;
	unsigned long ulStrength = (unsigned long)params[2];
	unsigned char port_no;
	unsigned long func_code;
	cell ret = -1;

	if( ( ( pf_code >> 8 ) & 0xFFFF )== 0 ){	/* w肳ꂽsɋ@\蓖ĂĂȂꍇ */
		if( g_NumDigitalOut < DIGITAL_IO_MAX ) {
			port_no = (pin_code >> 4) & 0x7;
			ulPort = g_ulGPIOBase[ port_no ];
			ucPins = ( 1 << (pin_code & 0x7) );
			func_code = PinFuncDigitalOut | g_NumDigitalOut;

			if( SetPinFuncCode( amx, pin_code , func_code ) == AMX_ERR_NONE )	/* IOR[hPawnVM֏߂@*/
			{
				MAP_SysCtlPeripheralEnable( g_ulSYSCTL_PERIPH_GPIO[ port_no ] );	/* w肵GPIOytFLɂ */
				MAP_GPIODirModeSet( ulPort , ucPins , GPIO_DIR_MODE_OUT);			/* w肵sGPIOo͂ƂĐݒ */
				MAP_GPIOPadConfigSet( ulPort , ucPins , ulStrength , GPIO_PIN_TYPE_STD );

				g_DO.Pins[ g_NumDigitalOut ].port = ulPort;
				g_DO.Pins[ g_NumDigitalOut ].pin  = ucPins;
				ret = (pin_code << 24) | func_code ;
				g_NumDigitalOut++;
			}
		}
	}

	return ret;
}

static cell AMX_NATIVE_CALL n_PinTypeGPIOInput(AMX *amx, const cell *params)
{
	unsigned long pf_code = ( unsigned long )params[1];
	unsigned char pin_code = ( unsigned char )( pf_code >> 24 );
	unsigned long ulPort;
	unsigned char ucPins;
	unsigned long ulPinType = (unsigned long)params[2];
	unsigned char port_no;
	unsigned long func_code;
	cell ret = -1;

	if( ( ( pf_code >> 8 ) & 0xFFFF )== 0 ){	/* w肳ꂽsɋ@\蓖ĂĂȂꍇ */
		if( g_NumDigitalIn < DIGITAL_IO_MAX ) {
			port_no = (pin_code >> 4) & 0x7;
			ulPort = g_ulGPIOBase[ port_no ];
			ucPins = ( 1 << (pin_code & 0x7) );
			func_code = PinFuncDigitalIn | g_NumDigitalIn;

			if( SetPinFuncCode( amx, pin_code , func_code ) == AMX_ERR_NONE )	/* IOR[hPawnVM֏߂@*/
			{
				MAP_SysCtlPeripheralEnable( g_ulSYSCTL_PERIPH_GPIO[ port_no ] );	/* w肵GPIOytFLɂ */
				MAP_GPIODirModeSet( ulPort , ucPins , GPIO_DIR_MODE_IN);			/* w肵sGPIO͂ƂĐݒ */
				MAP_GPIOPadConfigSet( ulPort , ucPins , GPIO_STRENGTH_2MA, ulPinType );
			
				g_DI.Pins[ g_NumDigitalIn ].port = ulPort;
				g_DI.Pins[ g_NumDigitalIn ].pin  = ucPins;
				ret = (pin_code << 24) | func_code ;
				g_NumDigitalIn++;
			}
		}
	}

	return ret;
}

static cell AMX_NATIVE_CALL n_GPIOWrite(AMX *amx,const cell *params)
{
	unsigned long pf_code = (unsigned long )params[1];
	unsigned char flag  = ( unsigned char )params[2];
	unsigned char dev_no;
	int done = -1;

	/* w肳ꂽs̏Ԃݒ */
	if( ( pf_code & PinFuncDigitalOut ) == PinFuncDigitalOut )
	{
		dev_no = pf_code & 0xF;
		if( dev_no < g_NumDigitalOut ) {
			if (flag)
			{
				MAP_GPIOPinWrite(g_DO.Pins[ dev_no ].port , g_DO.Pins[ dev_no ].pin , g_DO.Pins[dev_no].pin);
			}
			else
			{
				MAP_GPIOPinWrite(g_DO.Pins[ dev_no ].port , g_DO.Pins[ dev_no ].pin , 0);
			}
			done = 1;
		}
	}
	if( done == (-1) ) amx_putstr("illegal I/O\n");

	return 0;
}

static cell AMX_NATIVE_CALL n_GPIORead(AMX *amx,const cell *params)
{
	unsigned long pf_code = (unsigned long )params[1];
	unsigned char dev_no;
	cell ret = -1;

	/* w肳ꂽs̏Ԃǂݏo */
	dev_no = pf_code & 0xF;
	if( ( pf_code & PinFuncDigitalIn ) == PinFuncDigitalIn )	/* GPIOInput̏ꍇ */
	{
		if( dev_no < g_NumDigitalIn ) {
			ret = MAP_GPIOPinRead(g_DI.Pins[ dev_no ].port , g_DI.Pins[ dev_no ].pin) == g_DI.Pins[ dev_no ].pin;
		}
	}
	else if( ( pf_code & PinFuncDigitalOut ) == PinFuncDigitalOut )	/* GPIOOutput̏ꍇ */
	{
		if( dev_no < g_NumDigitalOut ) {
			ret = MAP_GPIOPinRead(g_DO.Pins[ dev_no ].port , g_DO.Pins[ dev_no ].pin) == g_DO.Pins[ dev_no ].pin;
		}
	}
	if( ret == (-1) ) amx_putstr("illegal I/O\n");

	return ret;
}

int amx_lm3s3748_GPIOInit(AMX *amx)
{
	static AMX_NATIVE_INFO MY_Extension[] = 
	{
		{ "PinTypeGPIOOutput" , n_PinTypeGPIOOutput },
		{ "PinTypeGPIOInput"  , n_PinTypeGPIOInput },
		{ "GPIOWrite"         , n_GPIOWrite },
		{ "GPIORead"          , n_GPIORead },
		{ NULL    , NULL }        // terminator
	};

	return amx_Register(amx , MY_Extension , -1);
}

int amx_lm3s3748_GPIOCleanup(AMX *amx){
	g_NumDigitalOut = 0;
	g_NumDigitalIn  = 0;

	return AMX_ERR_NONE;
}
