#include "inc/hw_memmap.h"
#include "inc/hw_types.h"

#include "driverlib/fpu.h"
#include "driverlib/debug.h"
#include "driverlib/gpio.h"
#include "driverlib/interrupt.h"
#include "driverlib/rom.h"
#include "driverlib/rom_map.h"
#include "driverlib/sysctl.h"
#include "driverlib/systick.h"

#include "inc/lm3s3748.h"

#include <stdio.h>
#include <string.h>
#include <alloca.h>

#include "usb_cdc_driver.h"
#include "amx.h"
#include "terminal.h"
#include "pawn_lm3s3748_IO.h"

#define VMAck       '\x01'
#define VMCSError   '\x02'
#define VMSizeError '\x03'
#define VMProgEnd   '\x04'
#define VMAbort     '\x05'
#define VMRTError   '\x06'

#define VMIdle              0x1
#define VMSizeOK            0x2
#define VMSizeCSError       0x3
#define VMByteCodeOK        0x4
#define VMByteCodeCSError   0x5
#define VMByteCodeError     0x6
#define VMRun               0x8
#define VMInterrupt         0x10

AMX g_amx;
static volatile tBoolean g_PAWNVMBreak = false;	/* IvtO */
int AMXAPI prun_Monitor(AMX *amx);				/* Ihook֐vg^Cv */

/* IvĎ֐ */
void USBCDCOptFunc()
{
	unsigned char cmd;
	unsigned long data_count;
	
	data_count = USBCDCBufferRead(&cmd, 1);
	if(data_count)
	{
		/* pwanVMsɏIvꍇ͂hook֐o^ */
		if( cmd == '\v' ){
			if(g_PAWNVMBreak == false) {
				g_PAWNVMBreak = true;
				amx_SetDebugHook(&g_amx, prun_Monitor);
			}
		}
	}
}

/* fBC֐[msP] */
void mwait (unsigned int pause_ms)
{
	while(pause_ms--)
	{
		MAP_SysCtlDelay( ( (MAP_SysCtlClockGet() / 3) / 1000 ) ); /* 1msx~ */
		if( g_PAWNVMBreak ) break;	/* Ivꍇ͔ */
	}
}

/* PawnVMpo͊֐ */
int amx_putstr(char* s) {
	int len = strlen(s);
	USBCDCBufferWrite( (unsigned char *)s, len);
	return len;
}

/* PawnVMpo͊֐ */
int amx_putchar(char c) {
	unsigned char outChar = c;
	USBCDCBufferWrite( &outChar, 1);
	return 1;
}

/* w肵[msP]x */
static cell AMX_NATIVE_CALL n_mwait(AMX *amx,const cell *params)
{
	unsigned int pause_ms = (unsigned int)params[1];
	mwait(pause_ms);
	return 0;
}


/* native֐Xg */
const AMX_NATIVE_INFO MY_Natives[] = 
{
	{ "printf"                  , n_printf },
	{ "wait_ms"                 , n_mwait },
	{ NULL , NULL }        // terminator
};

/* pwanVMIhook֐ */
int AMXAPI prun_Monitor(AMX *amx)
{
	return AMX_ERR_EXIT;
}

/* PawnVM{ */
int amx_proc(){
	int err;
	cell ret;
	AMX_HEADER hdr;
	unsigned char len_h , len_l , checkSUM;
	unsigned int sum , i;
	int  p_len ;
	unsigned char *prog_data;
	void *prog;
	unsigned char mesBuffer[3];
	unsigned long data_count;

	/* 3oCgM܂ő҂ */
	do{
		data_count = USBCDCBufferDataAvailable();
	}while(data_count < 3);
	
	USBCDCBufferRead(&len_l, 1);	/* oCgR[hiʃoCgjM */
	USBCDCBufferRead(&len_h, 1);	/* oCgR[hiʃoCgjM */
	USBCDCBufferRead(&checkSUM, 1);	/* `FbNTM */

	sum = (len_h + len_l) & 0xFF;	/* `FbNŤvZ */
	if( sum == checkSUM) {
		mesBuffer[0] = VMAck;		/* Ackݒ */
	}
	else {
		mesBuffer[0] = VMSizeCSError;	/* `FbNTG[ */
	}
	mesBuffer[1] = len_l;
	mesBuffer[2] = len_h;

	/* oCgR[h̑M */
	USBCDCBufferWrite( (unsigned char *)mesBuffer, 3);
	if(mesBuffer[0] != VMAck ) return VMSizeCSError;	/* G[ꍇ͔ */

	p_len = (len_h << 8) | len_l;	/* oCgR[ȟvZ */
	prog_data = alloca(p_len);		/* oCgR[hi[̈̊m */

	/* oCgR[h̎M */
	data_count=0;
	do
	{
		if(USBCDCBufferDataAvailable() > 0)
		{
			USBCDCBufferRead(&prog_data[data_count], 1);
			data_count++;
		}
	}while(data_count < p_len);

	/* `FbNTM */
	do{
		data_count = USBCDCBufferDataAvailable();
	}while(data_count < 1);
	USBCDCBufferRead(&checkSUM, 1);

	sum = 0;
	for(i=0 ; i< p_len ; i++) sum += prog_data[i];	/* `FbNŤvZ */
	if( (sum & 0xFF) == checkSUM) {
		mesBuffer[0] = VMAck;		/* Ackݒ */
	}
	else {
		mesBuffer[0] = VMSizeCSError;	/* `FbNTG[ */
	}
	
	/* 󋵂 */
	USBCDCBufferWrite( (unsigned char *)mesBuffer, 1);
	if(mesBuffer[0] != VMAck ) return VMByteCodeCSError;	/* G[ꍇ͔ */

	memcpy(&hdr, prog_data ,  sizeof(AMX_HEADER));	/* wb_[̎o */

	prog = alloca(hdr.stp);	/* VMp̍Ɨ̈̊m */
	memcpy(prog, prog_data ,  hdr.size);	/* vOf[^̃Rs[ */

	if (hdr.magic != AMX_MAGIC) {
		mesBuffer[0] = VMByteCodeError;		/* oCgR[hG[ */
	}
	else mesBuffer[0] = VMAck;		/* Ackݒ */
	
	mesBuffer[1] = ((int)hdr.stp) & 0xFF;		/* Ɨ̈TCYiʁj */
	mesBuffer[2] = ((int)hdr.stp >> 8) & 0xFF ;	/* Ɨ̈TCYiʁj */

	/* X^bNTCY𑗐M */
	USBCDCBufferWrite( (unsigned char *)mesBuffer, 3);
	if(mesBuffer[0] != VMAck ) return VMByteCodeError;	/* G[ꍇ͔ */

	memset(&g_amx, 0, sizeof(g_amx));
	err = amx_Init( &g_amx, prog );				/* VM̏ */
	err = amx_Register(&g_amx, MY_Natives, -1);	/* native֐̓o^ */
	amx_FloatInit(&g_amx);						/* FloatW[̓o^ */
	amx_LM3S3748_IOInit(&g_amx);				/* LM3S3748W[̓o^ */

	g_PAWNVMBreak = false;
	USBCDCIntRxExtRegister( USBCDCOptFunc );	/* IvĎ֐̓o^ */

	mwait(50);

	err = amx_Exec(&g_amx, &ret, AMX_EXEC_MAIN);	/* VM̎s */
	amx_LM3S3748_IOCleanup(&g_amx);					/* LM3S3748W[̔j */
	amx_Cleanup(&g_amx);							/* VM̔j */
	
	if (err != AMX_ERR_NONE){
		if(g_PAWNVMBreak)
		{
			mesBuffer[0] = VMAbort;		/* zXgAv̗vŉz}VBreak` */
			USBCDCBufferWrite( (unsigned char *)mesBuffer, 1);
		}
		else{
		/* runtimeG[̑M */
			mesBuffer[0] = VMRTError;	/* runntime error */
			mesBuffer[1] = err;			/* error nȏM       */
			mesBuffer[2] = VMRTError;	/* runntime error */
			USBCDCBufferWrite( (unsigned char *)mesBuffer, 3);
		}
	}
	else {
		mesBuffer[0] = VMProgEnd;		/* vOIƂzXgAvɓ` */
		USBCDCBufferWrite( (unsigned char *)mesBuffer, 1);
	}

	USBCDCIntRxExtUnregister();		/* IvĎ֐̎O */
	mwait(200);

	return err;
}

int main(void)
{
	/* VXeNbN50MHzɐݒ */
	SysCtlClockSet(SYSCTL_SYSDIV_4 | SYSCTL_USE_PLL | SYSCTL_OSC_MAIN | SYSCTL_XTAL_8MHZ);

	IntMasterEnable();

	/* CDC driveȑ */
	USBCDCInit();

	while(1) {
		amx_proc();
	}
}

